{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fgenerics #-}
{-# OPTIONS -fth #-}

module NominalAbs where

import Data.RepLib
import NominalBase
import NominalSubst

import List((\\),union)

class Nominal a b => BType a b where
    bound :: b -> [a]
    equiv :: b -> b -> Maybe (Perm a)

data Abs a b c = Abs b c deriving Show
$(derive [''Abs])


instance (Eq a, BType a b, Nominal a c) => Nominal a (Abs a b c) where
  swap t (Abs x y)           =  Abs (swap t x) (swap t y)
  fresh a (Abs x y)          =  a `elem` (bound x) || fresh a y
  aeq (p :: Proxy a) (Abs x y) (Abs x' y')  = (aeq p x x') && (aeq p y y')
				  || case ((equiv x x') :: Maybe (Perm a))
                                     of  Just pi -> all (\a -> fresh a y') ((bound x):: [a])
                                                    && aeq p y (perm pi y')
                                         Nothing -> False
  perm p (Abs x y)           = Abs (perm p x) (perm p y) 


unAbs :: (Rep1 (NominalD a) b, Rep1 (NominalD a) c, FreshMonad m a u, BType a b, Nominal a b, Nominal a c) => Abs a b c -> m (b, c)
unAbs ((Abs x y) :: Abs a b c) = let bs = (bound x :: [a]) in
	 	       do bs' <- mapM renameFM bs
			  return (perm (zip bs bs') (x,y))


instance (Rep1 (NominalD a) t, Rep1 (NominalD a) u, BType a t, Nominal a t, Nominal a u, Subst a v t,Subst a v u) => Subst a v (Abs a t u) where
    subst (v :: a)  t abs = do (x,y) <- unAbs abs; 
			       x' <- subst v t x;
			       y' <- subst v t y;
			       return (Abs x' y')


instance (Eq a, BType a t, Nominal a t, Nominal a u, FreeVars a s t, FreeVars a s u) => FreeVars a s (Abs a t u) where
    fvs dummy (Abs a x) = (fvs dummy a `union` fvs dummy x) \\ bound a



instance BType Name Name where
    bound a = [a]
    equiv a b = Just [(a,b)]


-- Binding 
-- 

data Binding n a = Binding n a deriving (Show, Eq)
$(derive [''Binding])

instance (Rep b, Rep n, BType a n, Eq b, Nominal a b) => BType a (Binding n b) where
    bound (Binding a x) = bound a
    equiv (Binding a x) (Binding b y) = if x == y then equiv a b else Nothing

   


