{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fallow-incoherent-instances #-}
{-# OPTIONS -fgenerics #-}
{-# OPTIONS -fth #-}

{- Nominal library base
   original - James Cheney
   mod to RepLib - Ben Kavanagh
-}

module NominalBase

where

import Data.RepLib
import Control.Monad.State

data Name = Name String (Maybe Int) deriving (Eq, Ord)
type Trans a = (a,a)
type Perm a = [Trans a]


-- derives R,R1 and Raux replib representations of Name. 
$(derive [''Name])

nmSep = ""

instance Show Name where
    showsPrec i (Name s (Just k)) = 
	showString s . showString nmSep . showsPrec i k
    showsPrec i (Name s Nothing) = 
	showString s


data Proxy a

data NominalD a b = NominalD {swapD :: Trans a -> b -> b,
			      freshD :: a -> b -> Bool,
			      aeqD :: Proxy a -> b -> b -> Bool,
			      permD :: Perm a -> b -> b}

instance Nominal a b => Sat (NominalD a b) where
    dict = NominalD swap fresh aeq perm


class  Nominal a b where
    swap  :: Trans a -> b -> b
    fresh :: a -> b -> Bool  
    aeq   :: Proxy a -> b -> b -> Bool
    perm  :: Perm a -> b -> b 


instance Rep1 (NominalD a) b => Nominal a b where
    swap  = swapR1 rep1
    fresh = freshR1 rep1
    aeq = aeqR1 rep1
    perm = permR1 rep1





applyMap :: R1 (NominalD a) b -> (forall b. Rep b => NominalD a b -> b -> b) -> b -> b
applyMap (Data1 dt cons) f x  =  
  case (findCon cons x) of 
      Val emb rec kids -> to emb $ map_l f rec kids  
applyMap _               _ x  = x


swapR1 :: R1 (NominalD a) b -> Trans a -> b -> b
swapR1 r1 t = applyMap r1 (\cb b -> swapD cb t b)


permR1 :: R1 (NominalD a) b -> Perm a -> b -> b
permR1 r1 p = applyMap r1 (\cb b -> permD cb p b)


freshR1 :: R1 (NominalD a) b -> a -> b -> Bool
freshR1 (Arrow1 r1 r2)  _ _ = error "nominal arrow urk"
freshR1 (Data1 dt cons) a x  =  
  case (findCon cons x) of 
      Val emb rec kids -> 
        foldl_l (\cb c b -> (freshD cb a b) && c) True rec kids
freshR1 _               _ _  = True


-- aeq almost identical to generic eq. (zipWith &&)
aeqR1 :: R1 (NominalD a) b -> Proxy a -> b -> b -> Bool
aeqR1 Int1 _           = (==)
aeqR1 Char1 _          = (==)
aeqR1 Integer1 _       = (==)      -- base types
aeqR1 Float1 _         = (==)
aeqR1 Double1 _        = (==)
aeqR1 Rational1 _      = (==) 
aeqR1 (Data1 _ cons) proxy = \x y -> 
   let loop (Con rcd rec : rest) = 
         case (from rcd x, from rcd y) of 
          (Just p1, Just p2) -> aeqRL1 rec p1 p2
          (Nothing, Nothing) -> loop rest
          (_,_) -> False
   in loop cons
aeqR1 r1 _ = error ("aeqR1 undefined for " ++ show r1)

aeqRL1 :: MTup (NominalD a) l -> l -> l -> Bool
aeqRL1 MNil Nil Nil = True
aeqRL1 (r :+: rl) (p1 :*: t1) (p2 :*: t2) =
  aeqD r undefined p1 p2 && aeqRL1 rl t1 t2




instance Eq a => Nominal a a where
    swap (a,b) c = if a == c then b 
		   else if b == c then a
		   else c
    perm pi a = foldr swap a pi
    fresh a b    = a /= b
    aeq _ x y = x == y


aeqN :: Nominal Name a => a -> a -> Bool
aeqN = aeq (undefined :: Proxy Name)


-- Freshness monad type class.

class NameChange n x | n -> x where 
    changeName :: n -> x -> n 

instance NameChange Name Int
    where
    changeName (Name s _) i = Name s (Just i)


class (Monad m) => UniqueSource m u where
    getUnique :: m u


-- FreshMonad is just a collection of constraints.
class (Monad m, NameChange n u, UniqueSource m u) => FreshMonad m n u | m -> u where
    renameFM     :: n -> m n
    renameFM n = do (u :: u) <- getUnique
		    return (changeName n u)


-- Simplest state example
type StateFM = State Int

instance UniqueSource StateFM Int where
    getUnique = do current <- get;
		   put (current + 1); 
		   return $ current


instance FreshMonad StateFM Name Int where

evalFM :: StateFM t -> t
evalFM fmcomp = evalState fmcomp 0




