{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fth #-}

{- Nominal library substitutions
   original - James Cheney
   Changed to using RepLib - Ben Kavanagh
-}

module NominalSubst where

import NominalBase
import Data.RepLib
import List(union)

class HasVar a b where
    is_var :: b -> Maybe a     -- retrieve the name of a variable
    var :: a -> b              -- inject name as a variable

instance HasVar Name Name where
    is_var a = Just a
    var a = a


-- substitute [a -> t] t'. 
class Subst a t t' where
    subst :: (FreshMonad m a u) => a -> t -> t' -> m t'

data SubstD a t t' = SubstD {substD:: (FreshMonad m a u) => a -> t -> t' -> m t'}

instance Subst a t t' => Sat (SubstD a t t') where
    dict = SubstD {substD = subst}

-- generic instance
instance Rep1 (SubstD a t) t' => Subst a t t' where
    subst = substR1 rep1

-- generic subst.
substR1 :: (Rep1 (SubstD a t) t', FreshMonad m a u) => R1 (SubstD a t) t' -> a -> t -> t' -> m t'
substR1 r (a::a) (t::t) t' = gmapM1 (\cb b -> substD cb a t b) t'

instance (Eq a, HasVar a t, Rep1 (SubstD a t) t) => Subst a t t where
    subst a t t' = if is_var t' == Just a 
		  then return t 
		  else gmapM1 (\cb b -> substD cb a t b) t'




-- all free variables of a particular type. 
-- requires dummy argument which represents the type of variables to collect

class FreeVars a t u where
    fvs :: t -> u -> [a]


data FreeVarsD a t u = FreeVarsD {fvsD::t -> u -> [a]}

instance FreeVars a t u => Sat (FreeVarsD a t u) where
    dict = FreeVarsD {fvsD = fvs}


-- I can either
-- (1) fold a gmapQ or 
-- (2) create a R1 function and call foldl_l
-- Here I use (1) 

instance (Eq a, Rep1 (FreeVarsD a t) u) => FreeVars a t u where
    fvs dummy t = foldl (List.union) [] (gmapQ1 (\cb b -> fvsD cb dummy b) t)
			  
instance (Eq a, HasVar a t, Rep1 (FreeVarsD a t) t) => FreeVars a t t where
   fvs dummy x = 
       case is_var x
       of Just a -> [a] 
	  Nothing -> foldl (List.union) [] (gmapQ1 (\cb b -> fvsD cb dummy b) x)


