{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fth #-}
{-# OPTIONS -fgenerics #-}
module PolyLam where

import NominalBase
import NominalAbs
import NominalSubst
import Data.RepLib
import List(find)
import Control.Monad.State
import Control.Monad.Reader

-- Second order lambda terms

data Type = VarTy Name
          | FnTy Type Type
          | AllTy (Abs Name Type) deriving Show

data Term = Var Name
          | App Term Term
          | Fn Type (Abs Name Term)
          | TyFn (Abs Name Term)
          | TyApp Term Type deriving Show

$(derive [''Type,''Term])

instance Eq Type where
    (==) = aeq (undefined :: Proxy Name)

instance Eq Term where
    (==) = aeq (undefined :: Proxy Name)

instance HasVar Name Type where
    var = VarTy
    is_var (VarTy x) = Just x
    is_var _ = Nothing

instance HasVar Name Term where
    var = Var
    is_var (Var x) = Just x
    is_var _ = Nothing

data TCEnv = TCEnv {tenv :: [Name], env :: [(Name,Type)]}
type TCM a = ReaderT TCEnv (State Int) a 

instance UniqueSource (ReaderT TCEnv (State Int)) Int where
    getUnique = do current <- get;
		   put (current + 1); 
		   return $ current

instance FreshMonad (ReaderT TCEnv (State Int)) Name Int

evalTCM m = evalState 0 $ runReaderT m (TCEnv [] [])

instance PolyTCM (ReaderT TCEnv (State Int)) Name Int where
    errorTC p = error 
    bindTV a m = local (\tc -> (TCEnv (a:tenv tc) (env tc))) m
    bindV a t m = local (\tc -> (TCEnv (tenv tc) ((a,t):env tc))) m
    lookupTV a  = ask >>= \tc -> return (a `elem` tenv tc)
    lookupV a   = ask >>= \tc -> 
		          case find (\(b,t) -> a == b) (env tc) 
			  of Just (a,t) -> return $ Just t
			     Nothing -> return $ Nothing

-- A freshness monad that also can store type and term variable bindings.
class FreshMonad m a u => PolyTCM m a u | m -> u where
    bindTV   ::  a -> m t -> m t
    bindV    ::  a -> Type -> m t -> m t
    lookupTV ::  a -> m Bool
    lookupV  ::  a -> m (Maybe Type)
    errorTC  ::  Proxy a -> String -> m t


-- two shortcut functions to avoid proxy args everywhere.
errorTCN :: PolyTCM m Name Int => String -> m t
errorTCN = errorTC (undefined :: Proxy Name)

unAbsN :: (Rep1 (NominalD Name) b, Rep1 (NominalD Name) c, FreshMonad m Name u, BType Name b, Nominal Name b, Nominal Name c) => Abs b c -> m (b, c)
unAbsN = unAbs (undefined :: Proxy Name)

wfTy                ::  PolyTCM m Name Int => Type -> m ()
wfTy (VarTy a)      =   do b <- lookupTV (a :: Name)
		           if b then return () else errorTCN "Unbound type variable"
wfTy (FnTy t1 t2)   =   do wfTy t1;
		           wfTy t2
wfTy (AllTy abs)    =   do (a,ty) <- unAbsN abs;
		           bindTV a (wfTy ty)

eqTy          ::  PolyTCM m Name Int => Type -> Type -> m ()
eqTy ty1 ty2  =   if ty1 == ty2 then return () else errorTCN "type mismatch"

matchFnTy                   ::  PolyTCM m Name Int => Type -> m (Type,Type)
matchFnTy (FnTy ty1 ty2)    =   return (ty1,ty2)
matchFnTy _                 =   errorTCN "Type mismatch: Expected function type"

matchAllTy              ::  PolyTCM m Name Int => Type -> m (Abs Name Type)
matchAllTy (AllTy abs)  =   return abs
matchAllTy _            =   errorTCN "Type mismatch: Expected forall type"


inferTm                       ::  PolyTCM m Name Int=> Term -> m Type
inferTm (Var x)               =   do  ty <- lookupV x
				      case ty 
				        of  Just ty' -> return ty'
				            Nothing -> errorTCN "Unbound variable"
inferTm (App t1 t2)           =   do  ty1 <- inferTm t1
				      (ty1',ty2') <- matchFnTy ty1
				      ty2 <- inferTm t2
				      eqTy ty1' ty2
				      return ty2'
inferTm (Fn ty abs)           =   do  (a,t) <- unAbsN abs
				      ty' <- bindV a ty (inferTm t)
				      return (FnTy ty ty')
				    
inferTm (TyApp tm ty)         =   do  ty' <- inferTm tm
				      wfTy ty
				      abs <- matchAllTy ty'
				      (a,ty'') <- unAbsN abs
				      subst a ty ty''
inferTm (TyFn abs)            =   do  (a,tm) <- unAbsN abs
				      ty <- bindTV a (inferTm tm)
				      return (AllTy (Abs a ty))
				    
		       

