{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
{-# OPTIONS -fth #-}
{-# OPTIONS -fgenerics #-}
module PolyLam2 where

import NominalBase
import NominalAbs
import NominalSubst
import Data.RepLib
import List(find)
import Control.Monad.State
import Control.Monad.Reader

-- Second order lambda terms

data NameTp = NameTp String (Maybe Int) deriving (Eq, Ord)
$(derive [''NameTp])

instance NameChange NameTp Int  where
    changeName (NameTp s _) i = NameTp s (Just i)

instance Show NameTp where
    showsPrec i (NameTp s n) = showsPrec i (Name s n)

instance BType NameTp NameTp where
    bound a = [a]
    equiv a b = Just [(a,b)]



data NameTm = NameTm String (Maybe Int) deriving (Eq, Ord)
$(derive [''NameTm])

instance NameChange NameTm Int where
    changeName (NameTm s _) i = NameTm s (Just i)

instance Show NameTm where
    showsPrec i (NameTm s n) = showsPrec i (Name s n)

instance BType NameTm NameTm where
    bound a = [a]
    equiv a b = Just [(a,b)]




aeqNTp :: Nominal NameTp a => a -> a -> Bool
aeqNTp = aeq (undefined :: Proxy NameTp)
aeqNTm :: Nominal NameTm a => a -> a -> Bool
aeqNTm = aeq (undefined :: Proxy NameTm)


data Type = VarTy NameTp
          | FnTy Type Type
          | AllTy (Abs NameTp NameTp Type) deriving Show

data Term = Var NameTm
          | App Term Term
          | Fn Type (Abs NameTm NameTm Term)
          | TyFn (Abs NameTp NameTp Term)
          | TyApp Term Type deriving Show

$(derive [''Type,''Term])

instance Eq Type where
    (==) = aeqNTp

instance Eq Term where
    (==) = aeqNTm

instance HasVar NameTp Type where
    var = VarTy
    is_var (VarTy x) = Just x
    is_var _ = Nothing

instance HasVar NameTm Term where
    var = Var
    is_var (Var x) = Just x
    is_var _ = Nothing

data TCEnv = TCEnv {tenv :: [NameTp], env :: [(NameTm,Type)]}
type TCM a = ReaderT TCEnv (State Int) a 

instance UniqueSource (ReaderT TCEnv (State Int)) Int where
    getUnique = do current <- get;
		   put (current + 1); 
		   return $ current

instance NameChange a Int => FreshMonad (ReaderT TCEnv (State Int)) a Int

evalTCM m = evalState 0 $ runReaderT m (TCEnv [] [])

instance PolyTCM (ReaderT TCEnv (State Int)) NameTp NameTm Int where
    errorTC p1 p2 = error 
    bindTV a p m = local (\tc -> (TCEnv (a:tenv tc) (env tc))) m
    bindV p a t m = local (\tc -> (TCEnv (tenv tc) ((a,t):env tc))) m
    lookupTV a p = ask >>= \tc -> return (a `elem` tenv tc)
    lookupV p a  = ask >>= \tc -> 
		          case find (\(b,t) -> a == b) (env tc) 
			  of Just (a,t) -> return $ Just t
			     Nothing -> return $ Nothing

-- A freshness monad that also can store type and term variable bindings.
class (FreshMonad m a u, FreshMonad m a' u) => PolyTCM m a a' u | m -> u where
    bindTV   ::  a -> Proxy a' -> m t -> m t
    bindV    ::  Proxy a -> a' -> Type -> m t -> m t
    lookupTV ::  a -> Proxy a' -> m Bool
    lookupV  ::  Proxy a -> a' -> m (Maybe Type)
    errorTC  ::  Proxy a -> Proxy a' -> String -> m t

bindTV'   ::  PolyTCM m NameTp NameTm Int => NameTp -> m t -> m t
bindTV' a = bindTV a (undefined :: Proxy NameTm)
bindV'    ::  PolyTCM m NameTp NameTm Int => NameTm -> Type -> m t -> m t
bindV' = bindV (undefined :: Proxy NameTp)
lookupTV' ::  PolyTCM m NameTp NameTm Int => NameTp -> m Bool
lookupTV' a = lookupTV a (undefined :: Proxy NameTm)
lookupV'  ::  PolyTCM m NameTp NameTm Int => NameTm -> m (Maybe Type)
lookupV' = lookupV (undefined :: Proxy NameTp)
errorTC' :: PolyTCM m NameTp NameTm Int => String -> m t
errorTC' = errorTC (undefined :: Proxy NameTp) (undefined :: Proxy NameTm)





unAbsTp :: (Rep1 (NominalD NameTp) b, Rep1 (NominalD NameTp) c, FreshMonad m NameTp u, BType NameTp b, Nominal NameTp b, Nominal NameTp c) => Abs NameTp b c -> m (b, c)
unAbsTp = unAbs 
unAbsTm :: (Rep1 (NominalD NameTm) b, Rep1 (NominalD NameTm) c, FreshMonad m NameTm u, BType NameTm b, Nominal NameTm b, Nominal NameTm c) => Abs NameTm b c -> m (b, c)
unAbsTm = unAbs 

wfTy                ::  PolyTCM m NameTp NameTm Int => Type -> m ()
wfTy (VarTy a)      =   do b <- lookupTV' (a :: NameTp)
		           if b then return () else errorTC' "Unbound type variable"
wfTy (FnTy t1 t2)   =   do wfTy t1;
		           wfTy t2
wfTy (AllTy abs)    =   do (a,ty) <- unAbsTp abs;
		           bindTV' a (wfTy ty)

eqTy          ::  PolyTCM m NameTp NameTm Int => Type -> Type -> m ()
eqTy ty1 ty2  =   if ty1 == ty2 then return () else errorTC' "type mismatch"

matchFnTy                   ::  PolyTCM m NameTp NameTm Int => Type -> m (Type,Type)
matchFnTy (FnTy ty1 ty2)    =   return (ty1,ty2)
matchFnTy _                 =   errorTC' "Type mismatch: Expected function type"

matchAllTy              ::  PolyTCM m NameTp NameTm Int => Type -> m (Abs NameTp NameTp Type)
matchAllTy (AllTy abs)  =   return abs
matchAllTy _            =   errorTC' "Type mismatch: Expected forall type"


inferTm                       ::  PolyTCM m NameTp NameTm Int=> Term -> m Type
inferTm (Var x)               =   do  ty <- lookupV' x
				      case ty 
				        of  Just ty' -> return ty'
				            Nothing -> errorTC' "Unbound variable"
inferTm (App t1 t2)           =   do  ty1 <- inferTm t1
				      (ty1',ty2') <- matchFnTy ty1
				      ty2 <- inferTm t2
				      eqTy ty1' ty2
				      return ty2'
inferTm (Fn ty abs)           =   do  (a,t) <- unAbsTm abs
				      ty' <- bindV' a ty (inferTm t)
				      return (FnTy ty ty')
				    
inferTm (TyApp tm ty)         =   do  ty' <- inferTm tm
				      wfTy ty
				      abs <- matchAllTy ty'
				      (a,ty'') <- unAbsTp abs
				      subst a ty ty''
inferTm (TyFn abs)            =   do  (a,tm) <- unAbsTp abs
				      ty <- bindTV' a (inferTm tm)
				      return (AllTy (Abs a ty))
				    
		       

